/******************************************************************************
 *                                                                            *
 *                                Copyright by                                *
 *                                                                            *
 *                              Azoteq (Pty) Ltd                              *
 *                          Republic of South Africa                          *
 *                                                                            *
 *                           Tel: +27(0)21 863 0033                           *
 *                           E-mail: info@azoteq.com                          *
 *                                                                            *
 * ========================================================================== *
 * @file        iqs7222b-example-code.ino                                     *
 * @brief       IQS7222B Twenty Button EV-Kit Example code                    *
 *              (PCB Version - AZP12193A3)                                    *
 *              This example demonstrates how to write the desired settings   *
 *              to the IQS7222B to use the Twenty Button EV-Kit.              *
 *                                                                            *
 *              Data is displayed over serial communication with the          *
 *              following outputs:                                            *
 *              - Touch detection on the 20 buttons (Channel 0 -> 19)         *
 *              - Power Mode Switching (if enabled)                           *
 *              - Force Communication (When 'f\n' is sent via Serial)         *
 *              - SW Reset (When 'r\n' is sent via Serial)                    *
 *                                                                            *
 *              When HID is enabled (DEMO_HID_ON 1), the keys will function   *
 *              as keyboard commands (Keep Serial Monitor open).              *
 *                                                                            *
 * @author      K. van der Walt - Azoteq PTY Ltd                              *
 * @version     v1.3                                                          *
 * @date        2023-07-10                                                    *
 *****************************************************************************/

#include <Arduino.h>
#include "src\IQS7222B\IQS7222B.h"

/*** Defines ***/
#define DEMO_IQS7222B_ADDR                     0x56
#define DEMO_IQS7222B_POWER_PIN                4
#define DEMO_IQS7222B_RDY_PIN                  7

/* Turn on HID and give keyboard commands over USB. */
#define DEMO_HID_ON 1

#if DEMO_HID_ON == 1
#include "Keyboard.h"
#endif

/*** Instances ***/
IQS7222B iqs7222b;

/*** Global Variables ***/
iqs7222b_ch_states ch_states[20] = {};
iqs7222b_power_modes power_mode = IQS7222B_NORMAL_POWER;
bool slider_event = false;
bool show_data = false;

void setup()
{
  /* Start Serial Communication */
  Serial.begin(115200);
  while(!Serial);
  Serial.println("Start Serial communication");
  delay(200);

  /* Power On IQS7222B */
  pinMode(DEMO_IQS7222B_POWER_PIN, OUTPUT);
  delay(200);
  digitalWrite(DEMO_IQS7222B_POWER_PIN, LOW);
  delay(200);
  digitalWrite(DEMO_IQS7222B_POWER_PIN, HIGH);

  /* Initialize the IQS7222B with input parameters device address and RDY pin */
  iqs7222b.begin(DEMO_IQS7222B_ADDR, DEMO_IQS7222B_RDY_PIN);
  Serial.println("IQS7222B Ready");
  delay(1);
}

void loop()
{
  iqs7222b.run(); // Runs the IQS7222B

  force_comms_and_read(); // function to initialize a force communication window.

  /* Process data read from IQS7222B when new data is available (RDY Line Low and
  new data were retrieved via I2C). */
  if(iqs7222b.new_data_available)
  {
    check_power_mode();
    check_channel_states();
    show_IQS7222B_data();
    iqs7222b.new_data_available = false;
  }
}

/* Function to check when the current power mode of the IQS7222B changed */
void check_power_mode(void)
{
  iqs7222b_power_modes buffer = iqs7222b.get_PowerMode(); // read current power mode
  if(buffer != power_mode)
  {
    power_mode = buffer;
    switch(power_mode)
    {
      case IQS7222B_NORMAL_POWER:
        Serial.println("NORMAL POWER Activated!");
        break;
      case IQS7222B_LOW_POWER:
        Serial.println("LOW POWER Activated!");
        break;
      case IQS7222B_ULP:
        Serial.println("ULP Activated!");
        break;
    }
  }
}

/* Function to check the touch states of all 20 mutual capacitance channels */
void check_channel_states(void)
{
  for(uint8_t i = 0; i < 20; i++)
  {
    /* check if the tou1ch state bit is set*/
    if(iqs7222b.channel_touchState((iqs7222b_channel_e)i))
    {
      if(ch_states[i] != IQS7222B_CH_TOUCH)
      {
        Serial.println("CH" + String(i) + ": Touch");
        ch_states[i] = IQS7222B_CH_TOUCH;
        #if DEMO_HID_ON
        keyboard_press((iqs7222b_channel_e)i);
        #endif
      }
    }
    /* NONE state if neither the touch nor proximity bit was set to 1 */
    else
    {
      if(ch_states[i] != IQS7222B_CH_NONE)
      {
        Serial.println("CH" + String(i) + ": None");
        ch_states[i] = IQS7222B_CH_NONE;
        #if DEMO_HID_ON
        Keyboard.releaseAll();
        #endif
      }
    }
  }
}

#if DEMO_HID_ON == 1
/* Function to write a specific keyboard key when HID is enabled */
void keyboard_press(iqs7222b_channel_e ch)
{
  switch (ch)
  {
    case IQS7222B_CH0:
      Keyboard.press(KEY_RIGHT_ARROW);
      break;

    case IQS7222B_CH1:
      Keyboard.press('#');
      break;

    case IQS7222B_CH2:
      Keyboard.press('4');
      break;

    case IQS7222B_CH3:
      Keyboard.press('D');
      break;

    case IQS7222B_CH4:
      Keyboard.press('8');
      break;

    case IQS7222B_CH5:
      Keyboard.press(KEY_DOWN_ARROW);
      break;

    case IQS7222B_CH6:
      Keyboard.press('0');
      break;

    case IQS7222B_CH7:
      Keyboard.press('3');
      break;

    case IQS7222B_CH8:
      Keyboard.press('C');
      break;

    case IQS7222B_CH9:
      Keyboard.press('7');
      break;

    case IQS7222B_CH10:
      Keyboard.press(KEY_UP_ARROW);
      break;

    case IQS7222B_CH11:
      Keyboard.press('9');
      break;

    case IQS7222B_CH12:
      Keyboard.press('2');
      break;

    case IQS7222B_CH13:
      Keyboard.press('B');
      break;

    case IQS7222B_CH14:
      Keyboard.press('6');
      break;

    case IQS7222B_CH15:
      Keyboard.press(KEY_LEFT_ARROW);
      break;

    case IQS7222B_CH16:
      Keyboard.press('*');
      break;

    case IQS7222B_CH17:
      Keyboard.press('1');
      break;

    case IQS7222B_CH18:
      Keyboard.press('A');
      break;

    case IQS7222B_CH19:
      Keyboard.press('5');
      break;
  }
}
#endif


/* Force the IQS7222B to open a RDY window and read the current state of the device */
void force_comms_and_read(void)
{
  char message = read_message();

  /* If an 'f' was received over serial, open a forced communication window and
  print the new data received */
  if(message == 'f')
  {
    iqs7222b.force_I2C_communication(); // prompt the IQS7222B
    show_data = true;
  }

  /* If an 'r' was received over serial, request a software(SW) reset */
  if(message == 'r')
  {
    Serial.println("Software Reset Requested!");
    iqs7222b.force_I2C_communication(); // Request a RDY window
    iqs7222b.iqs7222b_state.state = IQS7222B_STATE_SW_RESET;
    power_mode = IQS7222B_NORMAL_POWER;
  }
}

/* Read message sent over serial communication */
char read_message(void)
{
  while (Serial.available())
  {
    if(Serial.available() > 0)
    {
      return  Serial.read();
    }
  }

  return '\n';
}

/* shows the current IQS7222B data */
void show_IQS7222B_data()
{
  if(show_data)
  {
    Serial.println("***** IQS7222B DATA *****");
    Serial.println("1. All Channel States: ");
    for(uint8_t i = 0; i < 20; i++)
    {
      ch_states[i] = IQS7222B_CH_UNKNOWN;
    }
    check_channel_states();
    Serial.println("2. Power Mode:");
    power_mode = IQS7222B_POWER_UNKNOWN;
    check_power_mode();
    Serial.println("*************************");
    show_data = false;
  }
}
