/******************************************************************************
 *                                                                            *
 *                                Copyright by                                *
 *                                                                            *
 *                              Azoteq (Pty) Ltd                              *
 *                          Republic of South Africa                          *
 *                                                                            *
 *                           Tel: +27(0)21 863 0033                           *
 *                           E-mail: info@azoteq.com                          *
 *                                                                            *
 * ========================================================================== *
 * @file        IQS7222B.cpp                                                  *
 * @brief       This file contains the constructors and methods which allow   *
 *              ease of use of an IQS7222B Integrated Circuit (IC). The       *
 *              IQS7222B is a 20 Channel Mutual / 8 Channel Self-capacitive   *
 *              Touch and Proximity Controller with I2C communications        *
 *              interface, configurable GPIOs and low-power options.          *
 * @author      JN. Lochner - Azoteq PTY Ltd                                  *
 * @version     v1.1                                                          *
 * @date        2023                                                          *
 * ========================================================================== *
 * @attention  Makes use of the following standard Arduino libraries:         *
 * - Arduino.h -> Included in IQ7222B.h, comes standard with Arduino          *
 * - Wire.h    -> Included in IQS7222B.h, comes standard with Arduino         *
 *****************************************************************************/

/* Include Files */
#include "IQS7222B.h"
#include "IQS7222B_20_Button.h"

/* Private Global Variables */
bool iqs7222b_deviceRDY = false;
uint8_t iqs7222b_ready_pin;

/* Private Functions */
void iqs7222b_ready_interrupt(void);

/*****************************************************************************/
/*                               CONSTRUCTORS                                */
/*****************************************************************************/
IQS7222B::IQS7222B(){
}

/*****************************************************************************/
/*                              PUBLIC METHODS                               */
/*****************************************************************************/

/**
  * @name   begin
  * @brief  A method to initialize the IQS7222B device with the device address
  *         and ready pin specified by the user.
  * @param  deviceAddress ->  The address of the IQS7222B device.
  * @param  readyPin      ->  The Arduino pin which is connected to the ready
  *                           pin of the IQS7222B device.
  * @retval None.
  * @note   - Receiving a true return value does not mean that initialization
  *           was successful.
  *         - Receiving a true return value only means that the IQS device
  *           responded to the request for communication.
  *         - Receiving a false return value means that initialization
  *           did not take place at all.
  *         - If communication is successfully established then it is unlikely
  *           that initialization will fail.
  */
void IQS7222B::begin(uint8_t deviceAddressIn, uint8_t readyPinIn)
{
  /* Initialize I2C communication here, since this library can't function
  without it. */
  Wire.begin();
  Wire.setClock(400000);

  _deviceAddress = deviceAddressIn;
  iqs7222b_ready_pin = readyPinIn;
  attachInterrupt(digitalPinToInterrupt(iqs7222b_ready_pin), iqs7222b_ready_interrupt, CHANGE);

  /* Initialize "running" and "init" state machine variables. */
  iqs7222b_state.state      = IQS7222B_STATE_START;
  iqs7222b_state.init_state = IQS7222B_INIT_VERIFY_PRODUCT;
}

/**
  * @name   init
  * @brief  A method that runs through a normal start-up routine to set up the
  *         IQS7222B with the desired settings from the IQS7222B_init.h file.
  * @retval Returns true if the full start-up routine has been completed,
  *         returns false if not.
  * @param  None.
  * @note   No false return will be given, the program will thus be stuck when
  *         one of the cases is not able to finish. See serial communication to
  *         find the ERROR case
  */
bool IQS7222B::init(void)
{
  uint16_t prod_num;
  uint8_t ver_maj, ver_min;

  switch (iqs7222b_state.init_state)
  {
    /* Verifies product number to determine if the correct device is connected for
      this example */
    case IQS7222B_INIT_VERIFY_PRODUCT:
      if(iqs7222b_deviceRDY)
      {
        Serial.println("\tIQS7222B_INIT_VERIFY_PRODUCT");
        prod_num = getProductNum(RESTART);
        ver_maj = getmajorVersion(RESTART);
        ver_min = getminorVersion(STOP);
        Serial.print("\t\tProduct number is: ");
        Serial.print(prod_num);
        Serial.print(" v");
        Serial.print(ver_maj);
        Serial.print(".");
        Serial.println(ver_min);
        if(prod_num == IQS7222B_PRODUCT_NUM)
        {
          Serial.println("\t\tIQS7222B Release UI Confirmed!");
          iqs7222b_state.init_state = IQS7222B_INIT_READ_RESET;
        }
        else
        {
          Serial.println("\t\tDevice is not a IQS7222B!");
          iqs7222b_state.init_state = IQS7222B_INIT_NONE;
        }
      }
    break;

    /* Verify if a reset has occurred */
    case IQS7222B_INIT_READ_RESET:
      if(iqs7222b_deviceRDY)
      {
        Serial.println("\tIQS7222B_INIT_READ_RESET");
        updateInfoFlags(RESTART);
        if (checkReset())
        {
          Serial.println("\t\tReset event occurred.");
          iqs7222b_state.init_state = IQS7222B_INIT_UPDATE_SETTINGS;
        }
        else
        {
          Serial.println("\t\t No Reset Event Detected - Request SW Reset");
          iqs7222b_state.init_state = IQS7222B_INIT_CHIP_RESET;
        }
      }
    break;

    /* Perform SW Reset */
    case IQS7222B_INIT_CHIP_RESET:
      if(iqs7222b_deviceRDY)
      {
        Serial.println("\tIQS7222B_INIT_CHIP_RESET");

        //Perform SW Reset
        SW_Reset(STOP);
        Serial.println("\t\tSoftware Reset Bit Set.");
        delay(100);
        iqs7222b_state.init_state = IQS7222B_INIT_READ_RESET;
      }
    break;

    /* Write all settings to IQS7222B from .h file */
    case IQS7222B_INIT_UPDATE_SETTINGS:
      if(iqs7222b_deviceRDY)
      {
        Serial.println("\tIQS7222B_INIT_UPDATE_SETTINGS");
        writeMM(RESTART);
        iqs7222b_state.init_state = IQS7222B_INIT_ACK_RESET;
      }
    break;

    /* Acknowledge that the device went through a reset */
    case IQS7222B_INIT_ACK_RESET:
      if(iqs7222b_deviceRDY)
      {
        Serial.println("\tIQS7222B_INIT_ACK_RESET");
        acknowledgeReset(STOP);
        iqs7222b_state.init_state = IQS7222B_INIT_ATI;
      }
      break;

    /* Run the ATI algorithm to recalibrate the device with newly added settings */
    case IQS7222B_INIT_ATI:
      if(iqs7222b_deviceRDY)
      {
        Serial.println("\tIQS7222B_INIT_ATI");
        ReATI(STOP);
        iqs7222b_state.init_state = IQS7222B_INIT_WAIT_FOR_ATI;
        Serial.println("\tIQS7222B_INIT_WAIT_FOR_ATI");
      }
    break;

    /* Read the ATI Active bit to see if the rest of the program can continue */
    case IQS7222B_INIT_WAIT_FOR_ATI:
      if(iqs7222b_deviceRDY)
      {
        if(!readATIactive())
        {
          Serial.println("\t\tDONE");
          iqs7222b_state.init_state = IQS7222B_INIT_READ_DATA;
        }
      }
    break;

    /* Read the latest data from the iqs7222b */
    case IQS7222B_INIT_READ_DATA:
      if(iqs7222b_deviceRDY)
      {
        Serial.println("\tIQS7222B_INIT_READ_DATA");
        queueValueUpdates();
        iqs7222b_state.init_state = IQS7222B_INIT_ACTIVATE_EVENT_MODE;
      }
    break;

    /* Turn on I2C Stream-In-Touch mode */
    case IQS7222B_INIT_ACTIVATE_STREAM_IN_TOUCH_MODE:
      if(iqs7222b_deviceRDY)
      {
        Serial.println("\tIQS7222B_INIT_ACTIVATE_STREAM_IN_TOUCH_MODE");
        setStreamInTouchMode(STOP);
        iqs7222b_state.init_state = IQS7222B_INIT_DONE;
      }
    break;

    /* Turn on I2C Event mode */
    case IQS7222B_INIT_ACTIVATE_EVENT_MODE:
      if(iqs7222b_deviceRDY)
      {
        Serial.println("\tIQS7222B_INIT_ACTIVATE_EVENT_MODE");
        setEventMode(STOP);
        iqs7222b_state.init_state = IQS7222B_INIT_DONE;
      }
    break;

    /* If all operations has been completed correctly */
    case IQS7222B_INIT_DONE:
      Serial.println("\tIQS7222B_INIT_DONE");
      new_data_available = true;
      return true;
    break;

    default:
      break;
  }
  return false;

}

/**
  * @name   run
  * @brief  Run the device ready check and read the necessary bytes when the
  *         IQS7222B has pulled low the RDY line.
  *         The new_data_available flag will be set when a ready low is received
  *         from the IQS7222B.
  * @param  None
  * @retval None.
  * @note   queueValueUpdates can be edited by the user if other data should be
  *         read every time a RDY window is received.
  */
void IQS7222B::run(void)
{
  switch (iqs7222b_state.state)
  {
    /* After a hardware reset, this is the starting position of the main running 
      state machine */
    case IQS7222B_STATE_START:
      Serial.println("IQS7222B Initialization:");
      iqs7222b_state.state = IQS7222B_STATE_INIT;
    break;

    /* Perform the initialization routine on the IQS7222B */
    case IQS7222B_STATE_INIT:
      if(init())
      {
        Serial.println("IQS7222B Initialization complete!\n");
        iqs7222b_state.state = IQS7222B_STATE_RUN;
      }
    break;

    /* Send an 12C software reset in the next rdy window */
    case IQS7222B_STATE_SW_RESET:
      if(iqs7222b_deviceRDY)
      {
        SW_Reset(STOP);
        iqs7222b_state.state = IQS7222B_STATE_RUN;
      }
    break;

    /* Continuous reset monitoring state, ensure no reset event has occurred
      for data to be valid */
    case IQS7222B_STATE_CHECK_RESET:
      if(checkReset())
      {
        Serial.println("Reset Occurred!\n");
        new_data_available = false;
        iqs7222b_state.state = IQS7222B_STATE_START;
        iqs7222b_state.init_state = IQS7222B_INIT_VERIFY_PRODUCT;
      }

      /* A reset did not occur, move to the run state and wait for new ready
        window */
      else
      {
        new_data_available = true; /* No reset, thus data is valid */
        iqs7222b_state.state = IQS7222B_STATE_RUN;
      }
    break;

    /* If a RDY Window is open, read the latest values from the IQS7222B */
    case IQS7222B_STATE_RUN:
      if(iqs7222b_deviceRDY)
      {
        queueValueUpdates();
        iqs7222b_deviceRDY = false;
        new_data_available = false;
        iqs7222b_state.state = IQS7222B_STATE_CHECK_RESET;
      }
    break;
  }
}

/**
  * @name   iqs7222b_ready_interrupt
  * @brief  A method used as an interrupt function to set the iqs7222b_deviceRDY to
  *         the appropriate state. Activated when a High-Low and Low-High edge
  *         occurs.
  * @param  None.
  * @retval None.
  * @note   Keep this function as simple as possible to prevent stuck states
  *         and slow operations.
  */
void iqs7222b_ready_interrupt(void)
{
  if(digitalRead(iqs7222b_ready_pin))
  {
    iqs7222b_deviceRDY = false;
  }
  else
  {
    iqs7222b_deviceRDY = true;
  }
}

/**
  * @name   clearRDY
  * @brief  A method used to clear the ready interrupt bit.
  * @param  None.
  * @retval None.
  */
void IQS7222B::clearRDY(void)
{
  iqs7222b_deviceRDY = false;
}

/**
  * @name   getRDYStatus
  * @brief  A method used to retrieve the device RDY status.
  * @param  None.
  * @retval Returns the boolean IQS323 RDY state.
  *         - True when RDY line is LOW
  *         - False when RDY line is HIGH
  */
bool IQS7222B::getRDYStatus(void)
{
  return iqs7222b_deviceRDY;
}

/**
  * @name   queueValueUpdates
  * @brief  All I2C read operations in the queueValueUpdates method will be
  *         performed each time the IQS7222B opens a RDY window.
  * @param  None.
  * @retval None.
  * @note   Any Address in the memory map can be read from here.
  */
void IQS7222B::queueValueUpdates(void)
{
  uint8_t transferBytes[16];	// The array which will hold the bytes to be transferred.

	// Read the info flags, event flags, prox event states and touch event states.
	readRandomBytes(IQS7222B_MM_INFOFLAGS, 12, transferBytes, STOP);

  IQSMemoryMap.SYSTEM_STATUS[0]       = transferBytes[0];
  IQSMemoryMap.SYSTEM_STATUS[1]       = transferBytes[1];

  IQSMemoryMap.EVENTS[0]              = transferBytes[2];
  IQSMemoryMap.EVENTS[1]              = transferBytes[3];

  IQSMemoryMap.PROX_EVENT_STATES[0]   = transferBytes[4];
  IQSMemoryMap.PROX_EVENT_STATES[1]   = transferBytes[5];
  IQSMemoryMap.PROX_EVENT_STATES[2]   = transferBytes[6];
  IQSMemoryMap.PROX_EVENT_STATES[3]   = transferBytes[7];

  IQSMemoryMap.TOUCH_EVENT_STATES[0]  = transferBytes[8];
  IQSMemoryMap.TOUCH_EVENT_STATES[1]  = transferBytes[9];
  IQSMemoryMap.TOUCH_EVENT_STATES[2]  = transferBytes[10];
  IQSMemoryMap.TOUCH_EVENT_STATES[3]  = transferBytes[11];
}

/**
  * @name	  readATIactive
  * @brief  A method that checks if the ATI routine is still active
  * @param  None.
  * @retval Returns true if the ATI_ACTIVE_BIT is cleared, false if the
  *         ATI_ACTIVE_BIT is set.
  * @note   If the ATI routine is active the channel states (NONE, PROX, TOUCH)
  *         might exhibit unwanted behaviour. Thus it is advised to wait for
  *         the routine to complete before continuing.
  */
bool IQS7222B::readATIactive(void)
{
  /* Read the Info flags */
  updateInfoFlags(STOP);

  /* Return the ATI Active status */
  return getBit(IQSMemoryMap.SYSTEM_STATUS[0], IQS7222B_ATI_ACTIVE_BIT);
}

/**
  * @name	  checkReset
  * @brief  A method that checks if the device has reset and returns the reset 
  *         status.
  * @param  None.
  * @retval Returns true if a reset has occurred, false if no reset has occurred.
  * @note   If a reset has occurred the device settings should be reloaded using
  *         the begin function. After new device settings have been reloaded the
  *         acknowledge reset function can be used to clear the reset flag.
  */
bool IQS7222B::checkReset(void)
{
	/* Perform a bitwise AND operation inside getBIT with the DEVICE_RESET_BIT
  to return the reset status */
  return getBit(IQSMemoryMap.SYSTEM_STATUS[0], IQS7222B_DEVICE_RESET_BIT);
}

/**
  * @name	  checkProductNum
  * @brief  A method that checks the device product number and compares the
  *         result to the defined value to return a boolean result.
  * @param  stopOrRestart -> Specifies whether the communications window must
  *                          be kept open or must be closed after this action.
  *                          Use the STOP and RESTART definitions.
  * @retval Returns the product number as a 16-bit unsigned integer value.
  * @note   If the product is not correctly identified an appropriate messages
  *         should be displayed.
  */
uint16_t IQS7222B::getProductNum(bool stopOrRestart)
{
	uint8_t transferBytes[2];	      // A temporary array to hold the byte to be transferred.
  uint8_t prodNumLow = 0;         // Temporary storage for the Counts low byte.
  uint8_t prodNumHigh = 0;        // Temporary storage for the Counts high byte.
  uint16_t prodNumReturn = 0;     // The 16-bit return value.

	/* Read the Device info from the IQS7222B. */
	readRandomBytes(IQS7222B_MM_PROD_NUM, 2, transferBytes, stopOrRestart);

  /* Construct the 16-bit return value. */
  prodNumLow = transferBytes[0];
  prodNumHigh = transferBytes[1];
  prodNumReturn = (uint16_t)(prodNumLow);
  prodNumReturn |= (uint16_t)(prodNumHigh<<8);
  /* Return the counts value. */
  return prodNumReturn;
}

/**
  * @name	  getmajorVersion
  * @brief  A method that checks the device firmware version's major value.
  * @param  stopOrRestart -> Specifies whether the communications window must
  *                          be kept open or must be closed after this action.
  *                          Use the STOP and RESTART definitions.
  * @retval Returns major version number as an unsigned 8-bit integer value.
  */
uint8_t IQS7222B::getmajorVersion(bool stopOrRestart)
{
	uint8_t transferBytes[2];	// A temporary array to hold the byte to be transferred.
  uint8_t ver_maj = 0;      // Temporary storage for the firmware version major number.

	/* Read the Device info from the IQS7222B. */
	readRandomBytes(IQS7222B_MM_MAJOR_VERSION_NUM, 2, transferBytes, stopOrRestart);

  /* Get major value from correct byte */
  ver_maj = transferBytes[0];
  /* Return the major firmware version number value. */
  return ver_maj;
}

/**
  * @name	  getminorVersion
  * @brief  A method that checks the device firmware version's minor value.
  * @param  stopOrRestart -> Specifies whether the communications window must
  *                          be kept open or must be closed after this action.
  *                          Use the STOP and RESTART definitions.
  * @retval Returns minor version number as an unsigned 8-bit integer value.
  */
uint8_t IQS7222B::getminorVersion(bool stopOrRestart)
{
	uint8_t transferBytes[2];	// A temporary array to hold the byte to be transferred.
  uint8_t ver_min = 0;         // Temporary storage for the firmware version minor number.

	/* Read the Device info from the IQS7222B. */
	readRandomBytes(IQS7222B_MM_MINOR_VERSION_NUM, 2, transferBytes, stopOrRestart);
  /* get major value from correct byte */
  ver_min = transferBytes[0];
  /* Return the minor firmware version number value. */
  return ver_min;
}

/**
  * @name	  acknowledgeReset
  * @brief  A method that clears the Show Reset bit by writing a 1 to the
  *         IQS7222B_ACK_RESET_BIT   
  * @param  stopOrRestart -> Specifies whether the communications window must
  *                          be kept open or must be closed after this action.
  *                          Use the STOP and RESTART definitions.
  * @retval None.
  * @note   If a reset has occurred the device settings should be reloaded using
  *         the begin function. After new device settings have been reloaded this
  *         method should be used to clear the reset bit.
  */
void IQS7222B::acknowledgeReset(bool stopOrRestart)
{
	uint8_t transferByte[2];	// A temporary array to hold the bytes to be transferred.
	/* Read the System Flags from the IQS7222B, these must be read first in order
  not to change any settings. */
	readRandomBytes(IQS7222B_MM_CONTROL_SETTINGS, 2, transferByte, RESTART);
	/* Set the ACK_RESET_BIT in CONTROL_SETTINGS to clear the Reset Event Flag. */
	transferByte[0] = setBit(transferByte[0], IQS7222B_ACK_RESET_BIT);
	/* Write the new byte to the System Flags address. */
	writeRandomBytes(IQS7222B_MM_CONTROL_SETTINGS, 2, transferByte, stopOrRestart);
}

/**
  * @name   ReATI
  * @brief  A method which sets the REDO_ATI_BIT in order to force the IQS7222B
  *         device to run the Automatic Tuning Implementation (ATI) routine.
  * @param  stopOrRestart -> Specifies whether the communications window must
  *                          be kept open or must be closed after this action.
  *                          Use the STOP and RESTART definitions.
  * @retval None.
  * @note   To force ATI, RE_ATI_BIT in CONTROL_SETTINGS is set.
  */
void IQS7222B::ReATI(bool stopOrRestart)
{
  uint8_t transferByte[2]; // Array to store the bytes transferred.

  readRandomBytes(IQS7222B_MM_CONTROL_SETTINGS, 2, transferByte, RESTART);
  /* Set the RE_ATI_BIT in CONTROL_SETTINGS to force an ATI routine */
  transferByte[0] = setBit(transferByte[0], IQS7222B_RE_ATI_BIT);
  /* Write the new byte to the required device. */
  writeRandomBytes(IQS7222B_MM_CONTROL_SETTINGS, 2, transferByte, stopOrRestart);
}

/**
  * @name   SW_Reset
  * @brief  A method that sets the SW RESET bit to force the IQS7222B
  *         device to do a SW reset.
  * @param  stopOrRestart -> Specifies whether the communications window must
  *                          be kept open or must be closed after this action.
  *                          Use the STOP and RESTART definitions.
  * @retval None.
  * @note   To perform a SW Reset, SW_RESET_BIT in CONTROL_SETTINGS is set.
  */
void IQS7222B::SW_Reset(bool stopOrRestart)
{
  uint8_t transferByte[2]; // Array to store the bytes transferred.

  readRandomBytes(IQS7222B_MM_CONTROL_SETTINGS, 2, transferByte, RESTART);
  /* Set the SW_RESET_BIT in CONTROL_SETTINGS to perform a Software Reset */
  transferByte[0] = setBit(transferByte[0], IQS7222B_SW_RESET_BIT);
  /* Write the new byte to the required device. */
  writeRandomBytes(IQS7222B_MM_CONTROL_SETTINGS, 2, transferByte, stopOrRestart);
}

/**
  * @name   setStreamMode
  * @brief  A method to set the IQS7222B device into streaming mode.
  * @param  stopOrRestart ->  Specifies whether the communications window
  *                           must be kept open or must be closed after this action.
  *                           Use the STOP and RESTART definitions.
  * @retval None.
  * @note   All other bits at the IQS7222B_MM_CONTROL_SETTINGS register address 
  *         are preserved.
  */
void IQS7222B::setStreamMode(bool stopOrRestart)
{
  uint8_t transferBytes[2]; // The array which will hold the bytes which are transferred.

  // First read the bytes at the memory address so that they can be preserved.
  readRandomBytes(IQS7222B_MM_CONTROL_SETTINGS, 2, transferBytes, RESTART);
  // Set/Clear the INTERFACE_SELECTION_BITS in CONTROL_SETTINGS
  transferBytes[0] = clearBit(transferBytes[0], IQS7222B_INTERFACE_SELECT_BIT_0);
  transferBytes[0] = clearBit(transferBytes[0], IQS7222B_INTERFACE_SELECT_BIT_1);
  // Write the bytes back to the device
  writeRandomBytes(IQS7222B_MM_CONTROL_SETTINGS, 2, transferBytes, stopOrRestart);
}

/**
  * @name   setEventMode
  * @brief  A method to set the IQS7222B device into event mode.
  * @param  stopOrRestart ->  Specifies whether the communications window must
  *                           be kept open or must be closed after this action.
  *                           Use the STOP and RESTART definitions.
  * @retval None.
  * @note   All other bits at the IQS7222B_MM_CONTROL_SETTINGS register address 
  *         are preserved.
  */
void IQS7222B::setEventMode(bool stopOrRestart)
{
  uint8_t transferByte[2]; // The array which will hold the bytes which are transferred.

  /* First read the bytes at the memory address so that they can be preserved */
  readRandomBytes(IQS7222B_MM_CONTROL_SETTINGS, 2, transferByte, RESTART);
  /* Set/Clear the INTERFACE_SELECTION_BITS in CONTROL_SETTINGS */
  transferByte[0] = setBit(transferByte[0], IQS7222B_INTERFACE_SELECT_BIT_0);
  transferByte[0] = clearBit(transferByte[0], IQS7222B_INTERFACE_SELECT_BIT_1);
  /* Write the bytes back to the device */
  writeRandomBytes(IQS7222B_MM_CONTROL_SETTINGS, 2, transferByte, stopOrRestart);
}

/**
  * @name   setStreamInTouch
  * @brief  A method to set the IQS7222B device into streaming when in touch
  *         mode.
  * @param  stopOrRestart ->  Specifies whether the communications window must
  *                           be kept open or must be closed after this action.
  *                           Use the STOP and RESTART definitions.
  * @retval None.
  * @note   All other bits at the IQS7222B_MM_CONTROL_SETTINGS register address
  *         are preserved.
  */
void IQS7222B::setStreamInTouchMode(bool stopOrRestart)
{
  uint8_t transferBytes[2]; // The array which will hold the bytes which are transferred.

  // First read the bytes at the memory address so that they can be preserved.
  readRandomBytes(IQS7222B_MM_CONTROL_SETTINGS, 2, transferBytes, RESTART);
  // Set/Clear the INTERFACE_SELECTION_BITS in CONTROL_SETTINGS
  transferBytes[0] = clearBit(transferBytes[0], IQS7222B_INTERFACE_SELECT_BIT_0);
  transferBytes[0] = setBit(transferBytes[0], IQS7222B_INTERFACE_SELECT_BIT_1);
  // Write the bytes back to the device
  writeRandomBytes(IQS7222B_MM_CONTROL_SETTINGS, 2, transferBytes, stopOrRestart);
}

/**
  * @name   updateInfoFlags
  * @brief  A method that reads the IQS7222B info flags and assigns them to the
  *         SYSTEM_STATUS memory map local variable
  * @param  stopOrRestart ->  Specifies whether the communications window must
  *                           be kept open or must be closed after this action.
  *              			        Use the STOP and RESTART definitions.
  * @retval None.
  * @note   The SYSTEM_STATUS memory map local variable is altered with the new
  *         value of the info flags register retrieved from the IQS7222B.
  */
void IQS7222B::updateInfoFlags(bool stopOrRestart)
{
	uint8_t transferBytes[2];	// The array which will hold the bytes to be transferred.

	// Read the info flags.
	readRandomBytes(IQS7222B_MM_INFOFLAGS, 2, transferBytes, stopOrRestart);
	// Assign the info flags to the info flags union.
  IQSMemoryMap.SYSTEM_STATUS[0] =  transferBytes[0];
  IQSMemoryMap.SYSTEM_STATUS[1] =  transferBytes[1];
}

/**
  * @name   get_PowerMode
  * @brief  A method which reads the SYSTEM_STATUS memory map local variable
  *         and returns the current power mode.
  * @param  void
  * @retval Returns the current iqs7222b_power_modes state the device is in.
  * @note   See Datasheet on power mode options and timeouts.
  *         Normal Power, Low Power and Ultra Low Power (ULP).
  */
iqs7222b_power_modes IQS7222B::get_PowerMode(void)
{
  uint8_t buffer = getBit(IQSMemoryMap.SYSTEM_STATUS[0], IQS7222B_POWER_EVENT_BIT_0);
  buffer += getBit(IQSMemoryMap.SYSTEM_STATUS[0], IQS7222B_POWER_EVENT_BIT_1) << 1;

  if(buffer == IQS7222B_NORMAL_POWER_BIT)
  {
    return IQS7222B_NORMAL_POWER;
  }
  else if(buffer == IQS7222B_LOW_POWER_BIT)
  {
    return IQS7222B_LOW_POWER;
  }
  else if(buffer == IQS7222B_ULP_BIT)
  {
    return IQS7222B_ULP;
  }
  else
  {
    return IQS7222B_POWER_UNKNOWN;
  }
}

/**
  * @name   channel_touchState
  * @brief  A method that reads the unions for a given channel and determines
  *         if the channel is in a touch state.
  * @param  channel ->  The channel name on the IQS7222B (CH0-CH19) for which a
  *                     touch state needs to be determined.
  * @retval Returns true if a touch is active and false if there is no touch.
  * @note   See the iqs7222b_channel_e typedef for all possible channel names.
  */
bool IQS7222B::channel_touchState(iqs7222b_channel_e channel)
{
  switch(channel)
	{
		case IQS7222B_CH0:
			return getBit(IQSMemoryMap.TOUCH_EVENT_STATES[0], IQS7222B_CH0_TOUCH_BIT);
		break;

    case IQS7222B_CH1:
			return getBit(IQSMemoryMap.TOUCH_EVENT_STATES[0], IQS7222B_CH1_TOUCH_BIT);
		break;

    case IQS7222B_CH2:
			return getBit(IQSMemoryMap.TOUCH_EVENT_STATES[0], IQS7222B_CH2_TOUCH_BIT);
		break;

    case IQS7222B_CH3:
			return getBit(IQSMemoryMap.TOUCH_EVENT_STATES[0], IQS7222B_CH3_TOUCH_BIT);
		break;

    case IQS7222B_CH4:
			return getBit(IQSMemoryMap.TOUCH_EVENT_STATES[0], IQS7222B_CH4_TOUCH_BIT);
		break;

    case IQS7222B_CH5:
			return getBit(IQSMemoryMap.TOUCH_EVENT_STATES[0], IQS7222B_CH5_TOUCH_BIT);
		break;

    case IQS7222B_CH6:
			return getBit(IQSMemoryMap.TOUCH_EVENT_STATES[0], IQS7222B_CH6_TOUCH_BIT);
		break;

    case IQS7222B_CH7:
			return getBit(IQSMemoryMap.TOUCH_EVENT_STATES[0], IQS7222B_CH7_TOUCH_BIT);
		break;

    case IQS7222B_CH8:
			return getBit(IQSMemoryMap.TOUCH_EVENT_STATES[1], IQS7222B_CH8_TOUCH_BIT);
		break;

    case IQS7222B_CH9:
			return getBit(IQSMemoryMap.TOUCH_EVENT_STATES[1], IQS7222B_CH9_TOUCH_BIT);
		break;

    case IQS7222B_CH10:
			return getBit(IQSMemoryMap.TOUCH_EVENT_STATES[1], IQS7222B_CH10_TOUCH_BIT);
		break;

    case IQS7222B_CH11:
			return getBit(IQSMemoryMap.TOUCH_EVENT_STATES[1], IQS7222B_CH11_TOUCH_BIT);
		break;

    case IQS7222B_CH12:
			return getBit(IQSMemoryMap.TOUCH_EVENT_STATES[1], IQS7222B_CH12_TOUCH_BIT);
		break;

    case IQS7222B_CH13:
			return getBit(IQSMemoryMap.TOUCH_EVENT_STATES[1], IQS7222B_CH13_TOUCH_BIT);
		break;

    case IQS7222B_CH14:
			return getBit(IQSMemoryMap.TOUCH_EVENT_STATES[1], IQS7222B_CH14_TOUCH_BIT);
		break;

    case IQS7222B_CH15:
			return getBit(IQSMemoryMap.TOUCH_EVENT_STATES[1], IQS7222B_CH15_TOUCH_BIT);
		break;

    case IQS7222B_CH16:
			return getBit(IQSMemoryMap.TOUCH_EVENT_STATES[2], IQS7222B_CH16_TOUCH_BIT);
		break;

    case IQS7222B_CH17:
			return getBit(IQSMemoryMap.TOUCH_EVENT_STATES[2], IQS7222B_CH17_TOUCH_BIT);
		break;

    case IQS7222B_CH18:
			return getBit(IQSMemoryMap.TOUCH_EVENT_STATES[2], IQS7222B_CH18_TOUCH_BIT);
		break;

    case IQS7222B_CH19:
			return getBit(IQSMemoryMap.TOUCH_EVENT_STATES[2], IQS7222B_CH19_TOUCH_BIT);
		break;

		default:
			return false;
		break;
	}
}

/**
  * @name   channel_proxState
  * @brief  A method that reads the unions for a given channel and determines
  *         if the channel is in a proximity state.
  * @param  channel ->  The channel name on the IQS7222B (CH0-CH19) for which a
  *                     proximity state needs to be determined.
  * @retval Returns true if proximity is active and false if there is no proximity.
  * @note   See the iqs7222b_Channel_e typedef for all possible channel names.
  */
bool IQS7222B::channel_proxState(iqs7222b_channel_e channel)
{
  switch(channel)
	{
		case IQS7222B_CH0:
			return getBit(IQSMemoryMap.PROX_EVENT_STATES[0], IQS7222B_CH0_PROX_BIT);
		break;

    case IQS7222B_CH1:
			return getBit(IQSMemoryMap.PROX_EVENT_STATES[0], IQS7222B_CH1_PROX_BIT);
		break;

    case IQS7222B_CH2:
			return getBit(IQSMemoryMap.PROX_EVENT_STATES[0], IQS7222B_CH2_PROX_BIT);
		break;

    case IQS7222B_CH3:
			return getBit(IQSMemoryMap.PROX_EVENT_STATES[0], IQS7222B_CH3_PROX_BIT);
		break;

    case IQS7222B_CH4:
			return getBit(IQSMemoryMap.PROX_EVENT_STATES[0], IQS7222B_CH4_PROX_BIT);
		break;

    case IQS7222B_CH5:
			return getBit(IQSMemoryMap.PROX_EVENT_STATES[0], IQS7222B_CH5_PROX_BIT);
		break;

    case IQS7222B_CH6:
			return getBit(IQSMemoryMap.PROX_EVENT_STATES[0], IQS7222B_CH6_PROX_BIT);
		break;

    case IQS7222B_CH7:
			return getBit(IQSMemoryMap.PROX_EVENT_STATES[0], IQS7222B_CH7_PROX_BIT);
		break;

    case IQS7222B_CH8:
			return getBit(IQSMemoryMap.PROX_EVENT_STATES[1], IQS7222B_CH8_PROX_BIT);
		break;

    case IQS7222B_CH9:
			return getBit(IQSMemoryMap.PROX_EVENT_STATES[1], IQS7222B_CH9_PROX_BIT);
		break;

    case IQS7222B_CH10:
			return getBit(IQSMemoryMap.PROX_EVENT_STATES[1], IQS7222B_CH10_PROX_BIT);
		break;

    case IQS7222B_CH11:
			return getBit(IQSMemoryMap.PROX_EVENT_STATES[1], IQS7222B_CH11_PROX_BIT);
		break;

    case IQS7222B_CH12:
			return getBit(IQSMemoryMap.PROX_EVENT_STATES[1], IQS7222B_CH12_PROX_BIT);
		break;

    case IQS7222B_CH13:
			return getBit(IQSMemoryMap.PROX_EVENT_STATES[1], IQS7222B_CH13_PROX_BIT);
		break;

    case IQS7222B_CH14:
			return getBit(IQSMemoryMap.PROX_EVENT_STATES[1], IQS7222B_CH14_PROX_BIT);
		break;

    case IQS7222B_CH15:
			return getBit(IQSMemoryMap.PROX_EVENT_STATES[1], IQS7222B_CH15_PROX_BIT);
		break;

    case IQS7222B_CH16:
			return getBit(IQSMemoryMap.PROX_EVENT_STATES[2], IQS7222B_CH16_PROX_BIT);
		break;

    case IQS7222B_CH17:
			return getBit(IQSMemoryMap.PROX_EVENT_STATES[2], IQS7222B_CH17_PROX_BIT);
		break;

    case IQS7222B_CH18:
			return getBit(IQSMemoryMap.PROX_EVENT_STATES[2], IQS7222B_CH18_PROX_BIT);
		break;

    case IQS7222B_CH19:
			return getBit(IQSMemoryMap.PROX_EVENT_STATES[2], IQS7222B_CH19_PROX_BIT);
		break;

		default:
			return false;
		break;

	}
}
/*****************************************************************************/
/*										    	ADVANCED PUBLIC METHODS										       */
/*****************************************************************************/

/**
  * @name   writeMM
  * @brief  Function to write the whole memory map to the device (writable)
  *         registers
  * @param  stopOrRestart -> Specifies whether the communications window must
  *                          be kept open or must be closed after this action.
  *                          Use the STOP and RESTART definitions.
  * @retval None.
  * @note   IQS7222B_init.h -> exported GUI init.h file
  */
void IQS7222B::writeMM(bool stopOrRestart)
{
	uint8_t transferBytes[30];	// Temporary array which holds the bytes to be transferred.

  /* Change the Cycle Setup 0 - 4 */
  /* Memory Map Position 0x8000 - 0x8401 */
  transferBytes[0] = CYCLE_0_CONV_FREQ_FRAC;
  transferBytes[1] = CYCLE_0_CONV_FREQ_PERIOD;
  transferBytes[2] = CYCLE_0_SETTINGS;
  transferBytes[3] = CYCLE_0_CTX_SELECT;
  transferBytes[4] = CYCLE_1_CONV_FREQ_FRAC;
  transferBytes[5] = CYCLE_1_CONV_FREQ_PERIOD;
  transferBytes[6] = CYCLE_1_SETTINGS;
  transferBytes[7] = CYCLE_1_CTX_SELECT;
  transferBytes[8] = CYCLE_2_CONV_FREQ_FRAC;
  transferBytes[9] = CYCLE_2_CONV_FREQ_PERIOD;
  transferBytes[10] = CYCLE_2_SETTINGS;
  transferBytes[11] = CYCLE_2_CTX_SELECT;
  transferBytes[12] = CYCLE_3_CONV_FREQ_FRAC;
  transferBytes[13] = CYCLE_3_CONV_FREQ_PERIOD;
  transferBytes[14] = CYCLE_3_SETTINGS;
  transferBytes[15] = CYCLE_3_CTX_SELECT;
  transferBytes[16] = CYCLE_4_CONV_FREQ_FRAC;
  transferBytes[17] = CYCLE_4_CONV_FREQ_PERIOD;
  transferBytes[18] = CYCLE_4_SETTINGS;
  transferBytes[19] = CYCLE_4_CTX_SELECT;
  writeRandomBytes16(IQS7222B_MM_CYCLE_SETUP_0, 20, transferBytes, RESTART);
  Serial.println("\t\t1. Write Cycle 0-4 Settings");

  /* Change the Cycle Setup 5 - 9 */
  /* Memory Map Position 0x8500 - 0x8901 */
  transferBytes[0] = CYCLE_5_CONV_FREQ_FRAC;
  transferBytes[1] = CYCLE_5_CONV_FREQ_PERIOD;
  transferBytes[2] = CYCLE_5_SETTINGS;
  transferBytes[3] = CYCLE_5_CTX_SELECT;
  transferBytes[4] = CYCLE_6_CONV_FREQ_FRAC;
  transferBytes[5] = CYCLE_6_CONV_FREQ_PERIOD;
  transferBytes[6] = CYCLE_6_SETTINGS;
  transferBytes[7] = CYCLE_6_CTX_SELECT;
  transferBytes[8] = CYCLE_7_CONV_FREQ_FRAC;
  transferBytes[9] = CYCLE_7_CONV_FREQ_PERIOD;
  transferBytes[10] = CYCLE_7_SETTINGS;
  transferBytes[11] = CYCLE_7_CTX_SELECT;
  transferBytes[12] = CYCLE_8_CONV_FREQ_FRAC;
  transferBytes[13] = CYCLE_8_CONV_FREQ_PERIOD;
  transferBytes[14] = CYCLE_8_SETTINGS;
  transferBytes[15] = CYCLE_8_CTX_SELECT;
  transferBytes[16] = CYCLE_9_CONV_FREQ_FRAC;
  transferBytes[17] = CYCLE_9_CONV_FREQ_PERIOD;
  transferBytes[18] = CYCLE_9_SETTINGS;
  transferBytes[19] = CYCLE_9_CTX_SELECT;
  writeRandomBytes16(IQS7222B_MM_CYCLE_SETUP_5, 20, transferBytes, RESTART);
  Serial.println("\t\t2. Write Cycle 5-9 Settings");

  /* Change the Global Cycle Setup */
  /* Memory Map Position 0x8A00 - 0x8A02 */
  transferBytes[0] = GLOBAL_CYCLE_SETUP_0;
  transferBytes[1] = GLOBAL_CYCLE_SETUP_1;
  transferBytes[2] = COARSE_DIVIDER_PRELOAD;
  transferBytes[3] = FINE_DIVIDER_PRELOAD;
  transferBytes[4] = COMPENSATION_PRELOAD_0;
  transferBytes[5] = COMPENSATION_PRELOAD_1;
  writeRandomBytes16(IQS7222B_MM_GLOBAL_CYCLE_SETUP, 6, transferBytes, RESTART);
  Serial.println("\t\t3. Write Global Cycle Settings");

  /* Change the Button Setup 0 - 6 */
  /* Memory Map Position 0x9000 - 0x9601 */
  transferBytes[0] = BUTTON_0_PROX_THRESHOLD;
  transferBytes[1] = BUTTON_0_ENTER_EXIT;
  transferBytes[2] = BUTTON_0_TOUCH_THRESHOLD;
  transferBytes[3] = BUTTON_0_TOUCH_HYSTERESIS;
  transferBytes[4] = BUTTON_1_PROX_THRESHOLD;
  transferBytes[5] = BUTTON_1_ENTER_EXIT;
  transferBytes[6] = BUTTON_1_TOUCH_THRESHOLD;
  transferBytes[7] = BUTTON_1_TOUCH_HYSTERESIS;
  transferBytes[8] = BUTTON_2_PROX_THRESHOLD;
  transferBytes[9] = BUTTON_2_ENTER_EXIT;
  transferBytes[10] = BUTTON_2_TOUCH_THRESHOLD;
  transferBytes[11] = BUTTON_2_TOUCH_HYSTERESIS;
  transferBytes[12] = BUTTON_3_PROX_THRESHOLD;
  transferBytes[13] = BUTTON_3_ENTER_EXIT;
  transferBytes[14] = BUTTON_3_TOUCH_THRESHOLD;
  transferBytes[15] = BUTTON_3_TOUCH_HYSTERESIS;
  transferBytes[16] = BUTTON_4_PROX_THRESHOLD;
  transferBytes[17] = BUTTON_4_ENTER_EXIT;
  transferBytes[18] = BUTTON_4_TOUCH_THRESHOLD;
  transferBytes[19] = BUTTON_4_TOUCH_HYSTERESIS;
  transferBytes[20] = BUTTON_5_PROX_THRESHOLD;
  transferBytes[21] = BUTTON_5_ENTER_EXIT;
  transferBytes[22] = BUTTON_5_TOUCH_THRESHOLD;
  transferBytes[23] = BUTTON_5_TOUCH_HYSTERESIS;
  transferBytes[24] = BUTTON_6_PROX_THRESHOLD;
  transferBytes[25] = BUTTON_6_ENTER_EXIT;
  transferBytes[26] = BUTTON_6_TOUCH_THRESHOLD;
  transferBytes[27] = BUTTON_6_TOUCH_HYSTERESIS;
  writeRandomBytes16(IQS7222B_MM_BUTTON_SETUP_0, 28, transferBytes, RESTART);
  Serial.println("\t\t4. Write Button 0 - 6 Settings");

  /* Change the Button Setup 5 - 9 */
  /* Memory Map Position 0x9700 - 0x9D02 */
  transferBytes[0] = BUTTON_7_PROX_THRESHOLD;
  transferBytes[1] = BUTTON_7_ENTER_EXIT;
  transferBytes[2] = BUTTON_7_TOUCH_THRESHOLD;
  transferBytes[3] = BUTTON_7_TOUCH_HYSTERESIS;
  transferBytes[4] = BUTTON_8_PROX_THRESHOLD;
  transferBytes[5] = BUTTON_8_ENTER_EXIT;
  transferBytes[6] = BUTTON_8_TOUCH_THRESHOLD;
  transferBytes[7] = BUTTON_8_TOUCH_HYSTERESIS;
  transferBytes[8] = BUTTON_9_PROX_THRESHOLD;
  transferBytes[9] = BUTTON_9_ENTER_EXIT;
  transferBytes[10] = BUTTON_9_TOUCH_THRESHOLD;
  transferBytes[11] = BUTTON_9_TOUCH_HYSTERESIS;
  transferBytes[12] = BUTTON_10_PROX_THRESHOLD;
  transferBytes[13] = BUTTON_10_ENTER_EXIT;
  transferBytes[14] = BUTTON_10_TOUCH_THRESHOLD;
  transferBytes[15] = BUTTON_10_TOUCH_HYSTERESIS;
  transferBytes[16] = BUTTON_11_PROX_THRESHOLD;
  transferBytes[17] = BUTTON_11_ENTER_EXIT;
  transferBytes[18] = BUTTON_11_TOUCH_THRESHOLD;
  transferBytes[19] = BUTTON_11_TOUCH_HYSTERESIS;
  transferBytes[20] = BUTTON_12_PROX_THRESHOLD;
  transferBytes[21] = BUTTON_12_ENTER_EXIT;
  transferBytes[22] = BUTTON_12_TOUCH_THRESHOLD;
  transferBytes[23] = BUTTON_12_TOUCH_HYSTERESIS;
  transferBytes[24] = BUTTON_13_PROX_THRESHOLD;
  transferBytes[25] = BUTTON_13_ENTER_EXIT;
  transferBytes[26] = BUTTON_13_TOUCH_THRESHOLD;
  transferBytes[27] = BUTTON_13_TOUCH_HYSTERESIS;
  writeRandomBytes16(IQS7222B_MM_BUTTON_SETUP_7, 28, transferBytes, RESTART);
  Serial.println("\t\t5. Write Button 7 - 13 Settings");

  /* Change the Button Setup 14 - 19 */
  /* Memory Map Position 0x9E00 - 0xA302 */
  transferBytes[0] = BUTTON_14_PROX_THRESHOLD;
  transferBytes[1] = BUTTON_14_ENTER_EXIT;
  transferBytes[2] = BUTTON_14_TOUCH_THRESHOLD;
  transferBytes[3] = BUTTON_14_TOUCH_HYSTERESIS;
  transferBytes[4] = BUTTON_15_PROX_THRESHOLD;
  transferBytes[5] = BUTTON_15_ENTER_EXIT;
  transferBytes[6] = BUTTON_15_TOUCH_THRESHOLD;
  transferBytes[7] = BUTTON_15_TOUCH_HYSTERESIS;
  transferBytes[8] = BUTTON_16_PROX_THRESHOLD;
  transferBytes[9] = BUTTON_16_ENTER_EXIT;
  transferBytes[10] = BUTTON_16_TOUCH_THRESHOLD;
  transferBytes[11] = BUTTON_16_TOUCH_HYSTERESIS;
  transferBytes[12] = BUTTON_17_PROX_THRESHOLD;
  transferBytes[13] = BUTTON_17_ENTER_EXIT;
  transferBytes[14] = BUTTON_17_TOUCH_THRESHOLD;
  transferBytes[15] = BUTTON_17_TOUCH_HYSTERESIS;
  transferBytes[16] = BUTTON_18_PROX_THRESHOLD;
  transferBytes[17] = BUTTON_18_ENTER_EXIT;
  transferBytes[18] = BUTTON_18_TOUCH_THRESHOLD;
  transferBytes[19] = BUTTON_18_TOUCH_HYSTERESIS;
  transferBytes[20] = BUTTON_19_PROX_THRESHOLD;
  transferBytes[21] = BUTTON_19_ENTER_EXIT;
  transferBytes[22] = BUTTON_19_TOUCH_THRESHOLD;
  transferBytes[23] = BUTTON_19_TOUCH_HYSTERESIS;
  writeRandomBytes16(IQS7222B_MM_BUTTON_SETUP_14, 24, transferBytes, RESTART);
  Serial.println("\t\t6. Write Button 14 - 19 Settings");

  /* Change the CH0 - CH2 Setup */
  /* Memory Map Position 0xB000 - 0xB203 */
  transferBytes[0] = CH0_SETUP_0;
  transferBytes[1] = CH0_SETUP_1;
  transferBytes[2] = CH0_ATI_SETTINGS_0;
  transferBytes[3] = CH0_ATI_SETTINGS_1;
  transferBytes[4] = CH0_MULTIPLIERS_0;
  transferBytes[5] = CH0_MULTIPLIERS_1;
  transferBytes[6] = CH0_ATI_COMPENSATION_0;
  transferBytes[7] = CH0_ATI_COMPENSATION_1;
  transferBytes[8] = CH1_SETUP_0;
  transferBytes[9] = CH1_SETUP_1;
  transferBytes[10] = CH1_ATI_SETTINGS_0;
  transferBytes[11] = CH1_ATI_SETTINGS_1;
  transferBytes[12] = CH1_MULTIPLIERS_0;
  transferBytes[13] = CH1_MULTIPLIERS_1;
  transferBytes[14] = CH1_ATI_COMPENSATION_0;
  transferBytes[15] = CH1_ATI_COMPENSATION_1;
  transferBytes[16] = CH2_SETUP_0;
  transferBytes[17] = CH2_SETUP_1;
  transferBytes[18] = CH2_ATI_SETTINGS_0;
  transferBytes[19] = CH2_ATI_SETTINGS_1;
  transferBytes[20] = CH2_MULTIPLIERS_0;
  transferBytes[21] = CH2_MULTIPLIERS_1;
  transferBytes[22] = CH2_ATI_COMPENSATION_0;
  transferBytes[23] = CH2_ATI_COMPENSATION_1;
  writeRandomBytes16(IQS7222B_MM_CHANNEL_SETUP_0, 24, transferBytes, RESTART);
  Serial.println("\t\t7. Write Channel 0 - 2 Settings");

/* Change the CH3 - CH5 Setup */
/* Memory Map Position 0xB300 - 0xB503 */
  transferBytes[0] = CH3_SETUP_0;
  transferBytes[1] = CH3_SETUP_1;
  transferBytes[2] = CH3_ATI_SETTINGS_0;
  transferBytes[3] = CH3_ATI_SETTINGS_1;
  transferBytes[4] = CH3_MULTIPLIERS_0;
  transferBytes[5] = CH3_MULTIPLIERS_1;
  transferBytes[6] = CH3_ATI_COMPENSATION_0;
  transferBytes[7] = CH3_ATI_COMPENSATION_1;
  transferBytes[8] = CH4_SETUP_0;
  transferBytes[9] = CH4_SETUP_1;
  transferBytes[10] = CH4_ATI_SETTINGS_0;
  transferBytes[11] = CH4_ATI_SETTINGS_1;
  transferBytes[12] = CH4_MULTIPLIERS_0;
  transferBytes[13] = CH4_MULTIPLIERS_1;
  transferBytes[14] = CH4_ATI_COMPENSATION_0;
  transferBytes[15] = CH4_ATI_COMPENSATION_1;
  transferBytes[16] = CH5_SETUP_0;
  transferBytes[17] = CH5_SETUP_1;
  transferBytes[18] = CH5_ATI_SETTINGS_0;
  transferBytes[19] = CH5_ATI_SETTINGS_1;
  transferBytes[20] = CH5_MULTIPLIERS_0;
  transferBytes[21] = CH5_MULTIPLIERS_1;
  transferBytes[22] = CH5_ATI_COMPENSATION_0;
  transferBytes[23] = CH5_ATI_COMPENSATION_1;
  writeRandomBytes16(IQS7222B_MM_CHANNEL_SETUP_3, 24, transferBytes, RESTART);
  Serial.println("\t\t8. Write Channel 3 - 5 Settings");

  /* Change the CH6 - CH8 Setup */
  /* Memory Map Position 0xB600 - 0xB803 */
  transferBytes[0] = CH6_SETUP_0;
  transferBytes[1] = CH6_SETUP_1;
  transferBytes[2] = CH6_ATI_SETTINGS_0;
  transferBytes[3] = CH6_ATI_SETTINGS_1;
  transferBytes[4] = CH6_MULTIPLIERS_0;
  transferBytes[5] = CH6_MULTIPLIERS_1;
  transferBytes[6] = CH6_ATI_COMPENSATION_0;
  transferBytes[7] = CH6_ATI_COMPENSATION_1;
  transferBytes[8] = CH7_SETUP_0;
  transferBytes[9] = CH7_SETUP_1;
  transferBytes[10] = CH7_ATI_SETTINGS_0;
  transferBytes[11] = CH7_ATI_SETTINGS_1;
  transferBytes[12] = CH7_MULTIPLIERS_0;
  transferBytes[13] = CH7_MULTIPLIERS_1;
  transferBytes[14] = CH7_ATI_COMPENSATION_0;
  transferBytes[15] = CH7_ATI_COMPENSATION_1;
  transferBytes[16] = CH8_SETUP_0;
  transferBytes[17] = CH8_SETUP_1;
  transferBytes[18] = CH8_ATI_SETTINGS_0;
  transferBytes[19] = CH8_ATI_SETTINGS_1;
  transferBytes[20] = CH8_MULTIPLIERS_0;
  transferBytes[21] = CH8_MULTIPLIERS_1;
  transferBytes[22] = CH8_ATI_COMPENSATION_0;
  transferBytes[23] = CH8_ATI_COMPENSATION_1;
  writeRandomBytes16(IQS7222B_MM_CHANNEL_SETUP_6, 24, transferBytes, RESTART);
  Serial.println("\t\t9. Write Channel 6 - 8 Settings");

  /* Change the CH9 - CH11 Setup */
  /* Memory Map Position 0xB900 - 0xBB03 */
  transferBytes[0] = CH9_SETUP_0;
  transferBytes[1] = CH9_SETUP_1;
  transferBytes[2] = CH9_ATI_SETTINGS_0;
  transferBytes[3] = CH9_ATI_SETTINGS_1;
  transferBytes[4] = CH9_MULTIPLIERS_0;
  transferBytes[5] = CH9_MULTIPLIERS_1;
  transferBytes[6] = CH9_ATI_COMPENSATION_0;
  transferBytes[7] = CH9_ATI_COMPENSATION_1;
  transferBytes[8] = CH10_SETUP_0;
  transferBytes[9] = CH10_SETUP_1;
  transferBytes[10] = CH10_ATI_SETTINGS_0;
  transferBytes[11] = CH10_ATI_SETTINGS_1;
  transferBytes[12] = CH10_MULTIPLIERS_0;
  transferBytes[13] = CH10_MULTIPLIERS_1;
  transferBytes[14] = CH10_ATI_COMPENSATION_0;
  transferBytes[15] = CH10_ATI_COMPENSATION_1;
  transferBytes[16] = CH11_SETUP_0;
  transferBytes[17] = CH11_SETUP_1;
  transferBytes[18] = CH11_ATI_SETTINGS_0;
  transferBytes[19] = CH11_ATI_SETTINGS_1;
  transferBytes[20] = CH11_MULTIPLIERS_0;
  transferBytes[21] = CH11_MULTIPLIERS_1;
  transferBytes[22] = CH11_ATI_COMPENSATION_0;
  transferBytes[23] = CH11_ATI_COMPENSATION_1;
  writeRandomBytes16(IQS7222B_MM_CHANNEL_SETUP_9, 24, transferBytes, RESTART);
  Serial.println("\t\t10. Write Channel 9 - 11 Settings");

  /* Change the CH12 - CH14 Setup */
  /* Memory Map Position 0xBC00 - 0xBE03 */
  transferBytes[0] = CH12_SETUP_0;
  transferBytes[1] = CH12_SETUP_1;
  transferBytes[2] = CH12_ATI_SETTINGS_0;
  transferBytes[3] = CH12_ATI_SETTINGS_1;
  transferBytes[4] = CH12_MULTIPLIERS_0;
  transferBytes[5] = CH12_MULTIPLIERS_1;
  transferBytes[6] = CH12_ATI_COMPENSATION_0;
  transferBytes[7] = CH12_ATI_COMPENSATION_1;
  transferBytes[8] = CH13_SETUP_0;
  transferBytes[9] = CH13_SETUP_1;
  transferBytes[10] = CH13_ATI_SETTINGS_0;
  transferBytes[11] = CH13_ATI_SETTINGS_1;
  transferBytes[12] = CH13_MULTIPLIERS_0;
  transferBytes[13] = CH13_MULTIPLIERS_1;
  transferBytes[14] = CH13_ATI_COMPENSATION_0;
  transferBytes[15] = CH13_ATI_COMPENSATION_1;
  transferBytes[16] = CH14_SETUP_0;
  transferBytes[17] = CH14_SETUP_1;
  transferBytes[18] = CH14_ATI_SETTINGS_0;
  transferBytes[19] = CH14_ATI_SETTINGS_1;
  transferBytes[20] = CH14_MULTIPLIERS_0;
  transferBytes[21] = CH14_MULTIPLIERS_1;
  transferBytes[22] = CH14_ATI_COMPENSATION_0;
  transferBytes[23] = CH14_ATI_COMPENSATION_1;
  writeRandomBytes16(IQS7222B_MM_CHANNEL_SETUP_12, 24, transferBytes, RESTART);
  Serial.println("\t\t11. Write Channel 12 - 14 Settings");

  /* Change the CH15 - CH17 Setup */
  /* Memory Map Position 0xBF00 - 0xC103 */
  transferBytes[0] = CH15_SETUP_0;
  transferBytes[1] = CH15_SETUP_1;
  transferBytes[2] = CH15_ATI_SETTINGS_0;
  transferBytes[3] = CH15_ATI_SETTINGS_1;
  transferBytes[4] = CH15_MULTIPLIERS_0;
  transferBytes[5] = CH15_MULTIPLIERS_1;
  transferBytes[6] = CH15_ATI_COMPENSATION_0;
  transferBytes[7] = CH15_ATI_COMPENSATION_1;
  transferBytes[8] = CH16_SETUP_0;
  transferBytes[9] = CH16_SETUP_1;
  transferBytes[10] = CH16_ATI_SETTINGS_0;
  transferBytes[11] = CH16_ATI_SETTINGS_1;
  transferBytes[12] = CH16_MULTIPLIERS_0;
  transferBytes[13] = CH16_MULTIPLIERS_1;
  transferBytes[14] = CH16_ATI_COMPENSATION_0;
  transferBytes[15] = CH16_ATI_COMPENSATION_1;
  transferBytes[16] = CH17_SETUP_0;
  transferBytes[17] = CH17_SETUP_1;
  transferBytes[18] = CH17_ATI_SETTINGS_0;
  transferBytes[19] = CH17_ATI_SETTINGS_1;
  transferBytes[20] = CH17_MULTIPLIERS_0;
  transferBytes[21] = CH17_MULTIPLIERS_1;
  transferBytes[22] = CH17_ATI_COMPENSATION_0;
  transferBytes[23] = CH17_ATI_COMPENSATION_1;
  writeRandomBytes16(IQS7222B_MM_CHANNEL_SETUP_15, 24, transferBytes, RESTART);
  Serial.println("\t\t12. Write Channel 15 - 17 Settings");

/* Change the CH18 - CH19 Setup */
/* Memory Map Position 0xC200 - 0xC303 */
  transferBytes[0] = CH18_SETUP_0;
  transferBytes[1] = CH18_SETUP_1;
  transferBytes[2] = CH18_ATI_SETTINGS_0;
  transferBytes[3] = CH18_ATI_SETTINGS_1;
  transferBytes[4] = CH18_MULTIPLIERS_0;
  transferBytes[5] = CH18_MULTIPLIERS_1;
  transferBytes[6] = CH18_ATI_COMPENSATION_0;
  transferBytes[7] = CH18_ATI_COMPENSATION_1;
  transferBytes[8] = CH19_SETUP_0;
  transferBytes[9] = CH19_SETUP_1;
  transferBytes[10] = CH19_ATI_SETTINGS_0;
  transferBytes[11] = CH19_ATI_SETTINGS_1;
  transferBytes[12] = CH19_MULTIPLIERS_0;
  transferBytes[13] = CH19_MULTIPLIERS_1;
  transferBytes[14] = CH19_ATI_COMPENSATION_0;
  transferBytes[15] = CH19_ATI_COMPENSATION_1;
  writeRandomBytes16(IQS7222B_MM_CHANNEL_SETUP_18, 16, transferBytes, RESTART);
  Serial.println("\t\t13. Write Channel 18 - 19 Settings");

  /* Change the Filter Betas */
  /* Memory Map Position 0xAA00 - 0xAA01 */
  transferBytes[0] = COUNTS_BETA_FILTER;
  transferBytes[1] = LTA_BETA_FILTER;
  transferBytes[2] = LTA_FAST_BETA_FILTER;
  transferBytes[3] = RESERVED_FILTER_0;
  writeRandomBytes16(IQS7222B_MM_FILTER_BETAS, 4, transferBytes, RESTART);
  Serial.println("\t\t14. Write Filter Betas");

  /* Change the System Settings */
  /* Memory Map Position 0xD0 - 0xD9 */
  transferBytes[0] = SYSTEM_CONTROL_0;
  transferBytes[1] = SYSTEM_CONTROL_1;
  transferBytes[2] = ATI_ERROR_TIMEOUT_0;
  transferBytes[3] = ATI_ERROR_TIMEOUT_1;
  transferBytes[4] = ATI_REPORT_RATE_0;
  transferBytes[5] = ATI_REPORT_RATE_1;
  transferBytes[6] = NORMAL_MODE_TIMEOUT_0;
  transferBytes[7] = NORMAL_MODE_TIMEOUT_1;
  transferBytes[8] = NORMAL_MODE_REPORT_RATE_0;
  transferBytes[9] = NORMAL_MODE_REPORT_RATE_1;
  transferBytes[10] = LP_MODE_TIMEOUT_0;
  transferBytes[11] = LP_MODE_TIMEOUT_1;
  transferBytes[12] = LP_MODE_REPORT_RATE_0;
  transferBytes[13] = LP_MODE_REPORT_RATE_1;
  transferBytes[14] = ULP_MODE_TIMEOUT_0;
  transferBytes[15] = ULP_MODE_TIMEOUT_1;
  transferBytes[16] = ULP_MODE_REPORT_RATE_0;
  transferBytes[17] = ULP_MODE_REPORT_RATE_1;
  transferBytes[18] = TOUCH_PROX_EVENT_EN;
  transferBytes[19] = POWER_ATI_EVENT_EN;
  transferBytes[20] = I2CCOMMS_0;
  writeRandomBytes(IQS7222B_MM_CONTROL_SETTINGS, 21, transferBytes, stopOrRestart);
  Serial.println("\t\t15. Write System Settings");
}

/*****************************************************************************/
/*                            PRIVATE METHODS                                */
/*****************************************************************************/

/**
 * @name    readRandomBytes
 * @brief   A method that reads a specified number of bytes from a specified
 *          address and saves it into a user-supplied array.
 *          This method is used by all other methods in this class which read
 *          data from the IQS7222B device.
 * @param   memoryAddress -> The memory address at which to start reading bytes
 *          from.  See the "IQS7222B_addresses.h" file.
 * @param   numBytes      ->  The number of bytes that must be read.
 * @param   bytesArray    ->  The array which will store the bytes to be read,
 *                            this array will be overwritten.
 * @param   stopOrRestart ->  A boolean that specifies whether the communication
 *                            window should remain open or be closed after transfer.
 *                            False keeps it open, true closes it. Use the STOP
 *                            and RESTART definitions.
 * @retval  No value is returned, however, the user-supplied array is overwritten.
 * @note    Uses standard Arduino "Wire" library which is for I2C communication.
 *          Take note that C++ cannot return an array, therefore, the array which
 *          is passed as an argument is overwritten with the required values.
 *          Pass an array to the method by using only its name, e.g. "bytesArray",
 *          without the brackets, this basically passes a pointer to the array.
 */
void IQS7222B::readRandomBytes(uint8_t memoryAddress, uint8_t numBytes, uint8_t bytesArray[], bool stopOrRestart)
{
  // A simple counter to assist with loading bytes into the user-supplied array.
  uint8_t i = 0;
  // Select the device with the address of "_deviceAddress" and start communication.
  Wire.beginTransmission(_deviceAddress);
  // Send a bit asking for the "memoryAddress" register.
  Wire.write(memoryAddress);
  // Complete the selection and communication initialization.
  Wire.endTransmission(RESTART);  // Restart transmission for reading that follows.
  /* The required device has now been selected and it has been told which register
  to send information from. */

  // Request "numBytes" bytes from the device which has the address: "_deviceAddress"
  do
  {
    Wire.requestFrom((int)_deviceAddress, (int)numBytes, (int)stopOrRestart);
  }while(Wire.available() == 0);  // Wait for response, this sometimes takes a few attempts

  // Load the received bytes into the array until there are no more
  while(Wire.available())
  {
    // Load the received bytes into the user-supplied array
    bytesArray[i] = Wire.read();
    i++;
  }

  /* Always manually close the RDY window after a STOP is sent to prevent
  writing while the RDY window closes */
  if(stopOrRestart == STOP)
  {
    iqs7222b_deviceRDY = false;
  }
}

/**
  * @name   writeRandomBytes
  * @brief  A method that writes a specified number of bytes to a specified
  *         address, the bytes to write are supplied using an array
  *         pointer. This method is used by all other methods of this
  *         class which writes data to the IQS7222B device.
  * @param  memoryAddress ->  The memory address at which to start writing the
  *                           bytes to. See the "IQS7222B_addresses.h" file.
  * @param  numBytes      ->  The number of bytes that must be written.
  * @param  bytesArray    ->  The array which stores the bytes which will be
  *                           written to the memory location.
  * @param  stopOrRestart ->  A boolean that specifies whether the communication
  *                           window should remain open or be closed of transfer.
  *                           False keeps it open, true closes it. Use the STOP
  *                           and RESTART definitions.
  * @retval No value is returned, only the IQS device registers are altered.
  * @note   Uses standard Arduino "Wire" library which is for I2C communication.
  *         Take note that a full array cannot be passed to a function in C++.
  *         Pass an array to the function by using only its name, e.g. "bytesArray",
  *         without the square brackets, this passes a pointer to the
  *         array. The values to be written must be loaded into the array prior
  *         to passing it to the function.
  */
void IQS7222B::writeRandomBytes(uint8_t memoryAddress, uint8_t numBytes, uint8_t bytesArray[], bool stopOrRestart)
{
  // Select the device with the address of "_deviceAddress" and start communication.
  Wire.beginTransmission(_deviceAddress);
  /* Specify the memory address where the IQS7222B must start saving the data,
  as designated by the "memoryAddress" variable. */
  Wire.write(memoryAddress);
  /* Write the bytes as specified in the array that "arrayAddress" pointer
  points to. */
  for(int i=0; i<numBytes; i++)
  {
    Wire.write(bytesArray[i]);
  }
  // End the transmission, user decides to STOP or RESTART.
  Wire.endTransmission(stopOrRestart);

  /* Always manually close the RDY window after a STOP is sent to prevent
  writing while the RDY window closes */
  if(stopOrRestart == STOP)
  {
    iqs7222b_deviceRDY = false;
  }
}

/**
  * @name   writeRandomBytes
  * @brief  A method that writes a specified number of bytes to a specified
  *         address, the bytes to write are supplied using an array pointer.
  *         This method is used by all other methods of this class which
  *         write data to the IQS7222B device.
  * @param  memoryAddress ->  The memory address at which to start writing the
  *                           bytes to. See the "IQS7222B_addresses.h" file.
  * @param  numBytes      ->  The number of bytes that must be written.
  * @param  bytesArray    ->  The array which stores the bytes which will be
  *                           written to the memory location.
  * @param  stopOrRestart ->  A boolean that specifies whether the communication
  *                           window should remain open or be closed of transfer.
  *                           False keeps it open, true closes it. Use the STOP
  *                           and RESTART definitions.
  * @retval No value is returned, only the IQS device registers are altered.
  * @note   Uses standard Arduino "Wire" library which is for I2C communication.
  *         Take note that a full array cannot be passed to a function in C++.
  *         Pass an array to the function by using only its name, e.g. "bytesArray",
  *         without the square brackets, this passes a pointer to the
  *         array. The values to be written must be loaded into the array prior
  *         to passing it to the function.
  */
void IQS7222B::writeRandomBytes16(uint16_t memoryAddress, uint8_t numBytes, uint8_t bytesArray[], bool stopOrRestart)
{
  // Select the device with the address of "_deviceAddress" and start communication.
  Wire.beginTransmission(_deviceAddress);
  /* Specify the memory address where the IQS7222B must start saving the data,
  as designated by the "memoryAddress" variable. */
  uint8_t addr_h, addr_l;
  addr_h = memoryAddress >> 8;
  addr_l = memoryAddress;
  Wire.write(addr_h);
  Wire.write(addr_l);
  /* Write the bytes as specified in the array that "arrayAddress" pointer
  points to. */
  for(int i=0; i<numBytes; i++)
  {
    Wire.write(bytesArray[i]);
  }
  // End the transmission, user decides to STOP or RESTART.
  Wire.endTransmission(stopOrRestart);

  /* Always manually close the RDY window after a STOP is sent to prevent
  writing while the RDY window closes */
  if(stopOrRestart == STOP)
  {
    iqs7222b_deviceRDY = false;
  }
}

/**
  * @name   getBit
  * @brief  A method which returns the chosen bit value of the provided byte.
  * @param  data       -> byte of which a given bit value needs to be calculated.
  * @param  bit_number -> a number between 0 and 7 representing the bit in question.
  * @retval The boolean value of the specific bit requested. 
  */
bool IQS7222B::getBit(uint8_t data, uint8_t bit_number)
{
  return (data & ( 1 << bit_number )) >> bit_number;
}

/**
  * @name   setBit
  * @brief  A method which returns the chosen bit value of the provided byte.
  * @param  data       -> byte of which a given bit value needs to be calculated.
  * @param  bit_number -> a number between 0 and 7 representing the bit in question.
  * @retval Returns an 8-bit unsigned integer value of the given data byte with 
  *         the requested bit set.
  */
uint8_t IQS7222B::setBit(uint8_t data, uint8_t bit_number)
{
	return (data |= 1UL << bit_number);
}

/**
  * @name   clearBit
  * @brief  A method which returns the chosen bit value of the provided byte.
  * @param  data       -> byte of which a given bit value needs to be calculated.
  * @param  bit_number -> a number between 0 and 7 representing the bit in question.
  * @retval Returns an 8-bit unsigned integer value of the given data byte with 
  *         the requested bit cleared.
  */
uint8_t IQS7222B::clearBit(uint8_t data, uint8_t bit_number)
{
	return (data &= ~(1UL << bit_number));
}

/**
  * @name   force_I2C_communication
  * @brief  A method which writes to memory address 0xFF to open a
  *         communication window on the IQS7222B.
  * @param  None
  * @retval None
  * @note   Uses standard Arduino "Wire" library which is for I2C communication.
  */
void IQS7222B::force_I2C_communication(void)
{
  /*Ensure RDY is HIGH at the moment*/
  if (!iqs7222b_deviceRDY)
  {
    /* Select the device with the address of "DEMO_IQS7222B_ADDR" and start
    communication. */
    Wire.beginTransmission(_deviceAddress);

    /* Write to memory address 0xFF that will prompt the IQS7222B to open a
    communication window.*/
    Wire.write(0xFF);

    /* End the transmission, user decides to STOP or RESTART. */
    Wire.endTransmission(STOP);
  }
}
